@@ -357,6 +357,8 @@ pub struct GridAxes {
     pub x_grid: Vec<f64>,
     /// Parton IDs used in the grid.
     pub pids: Vec<i32>,
+    /// Interpolation grid for the renormalization scale of the `Grid`.
+    pub mur2_grid: Vec<f64>,
     /// Interpolation grid for the factorization scale of the `Grid`.
     pub muf2_grid: Vec<f64>,
 }
@@ -366,7 +368,7 @@ pub struct GridAxes {
 pub struct EkoInfo {
     /// Scale of the FkTable.
     pub muf2_0: f64,
-    /// Strong coupling constants for the factorization scales in the same ordering as given in
+    /// Strong coupling constants for the renormalization scales in the same ordering as given in
     /// [`GridAxes`].
     pub alphas: Vec<f64>,
     /// Renormalization scale variation.
@@ -572,8 +574,6 @@ impl Grid {
         bins
     }
 
-    // TODO: migrate to use a LumiCache
-
     /// Convolutes a single subgrid `(order, bin, lumi)` with the PDFs strong coupling given by
     /// `xfx1`, `xfx2` and `alphas`. The convolution result is fully differentially, such that the
     /// axes of the result correspond to the values given by the subgrid `q2`, `x1` and `x2` grid
@@ -724,6 +724,10 @@ impl Grid {
     /// # Errors
     ///
     /// If writing or compression fails an error is returned.
+    ///
+    /// # Panics
+    ///
+    /// TODO
     pub fn write_lz4(&self, writer: impl Write) -> Result<(), GridError> {
         let mut encoder = FrameEncoder::new(writer);
         self.write(&mut encoder)?;
@@ -970,12 +974,28 @@ impl Grid {
         }
     }
 
+    /// Scales each subgrid by a bin-dependent factor given in `factors`. If a bin does not have a
+    /// corresponding entry in `factors` it is not rescaled. If `factors` has more entries than
+    /// there are bins the superfluous entries do not have an effect.
+    pub fn scale_by_bin(&mut self, factors: &[f64]) {
+        for ((_, bin, _), subgrid) in self.subgrids.indexed_iter_mut() {
+            if let Some(&factor) = factors.get(bin) {
+                subgrid.scale(factor);
+            }
+        }
+    }
+
     /// Returns the subgrid parameters.
     #[must_use]
     pub fn orders(&self) -> &[Order] {
         &self.orders
     }
 
+    /// Set the luminosity function for this grid.
+    pub fn set_lumis(&mut self, lumis: Vec<LumiEntry>) {
+        self.lumi = lumis;
+    }
+
     /// Returns the subgrid with the specified indices `order`, `bin`, and `lumi`.
     #[must_use]
     pub fn subgrid(&self, order: usize, bin: usize, lumi: usize) -> &SubgridEnum {
@@ -1271,6 +1291,10 @@ impl Grid {
 
     /// Provide information used to compute a suitable EKO for the current grid.
     /// More specific, the `x_grid` and `muf2_grid` are extracted and checked.
+    ///
+    /// # Panics
+    ///
+    /// TODO
     #[must_use]
     pub fn axes(&self) -> Option<GridAxes> {
         // determine what and how many hadrons are in the initial state
@@ -1299,6 +1323,7 @@ impl Grid {
             .iter()
             .all(|entry| entry.entry().iter().all(|&(_, b, _)| b == initial_state_2));
 
+        let mut mur2_grid = Vec::new();
         let mut muf2_grid = Vec::new();
         let mut x_grid = Vec::new();
         let pids = Vec::new();
@@ -1338,6 +1363,7 @@ impl Grid {
 
             // not all luminosities are equal (some appear only at higher orders)
             for subgrid in lane.iter() {
+                mur2_grid.append(&mut subgrid.mu2_grid().iter().map(|mu2| mu2.ren).collect());
                 muf2_grid.append(&mut subgrid.mu2_grid().iter().map(|mu2| mu2.fac).collect());
                 if has_pdf1 {
                     x_grid.extend_from_slice(&subgrid.x1_grid());
@@ -1348,14 +1374,18 @@ impl Grid {
             }
         }
 
-        muf2_grid.sort_by(|a, b| a.partial_cmp(b).unwrap_or_else(|| unreachable!()));
-        muf2_grid.dedup_by(|a, b| approx_eq!(f64, *a, *b, ulps = 64));
-        x_grid.sort_by(|a, b| a.partial_cmp(b).unwrap_or_else(|| unreachable!()));
+        // make grids unique
+        x_grid.sort_by(|a, b| a.partial_cmp(b).unwrap());
         x_grid.dedup_by(|a, b| approx_eq!(f64, *a, *b, ulps = 64));
+        mur2_grid.sort_by(|a, b| a.partial_cmp(b).unwrap());
+        mur2_grid.dedup_by(|a, b| approx_eq!(f64, *a, *b, ulps = 64));
+        muf2_grid.sort_by(|a, b| a.partial_cmp(b).unwrap());
+        muf2_grid.dedup_by(|a, b| approx_eq!(f64, *a, *b, ulps = 64));
 
         Some(GridAxes {
             x_grid,
             pids, // TODO: for the time being they are just empty, but we might use them for slicing the eko
+            mur2_grid,
             muf2_grid,
         })
     }
@@ -1378,6 +1408,7 @@ impl Grid {
         // Check operator layout
         let dim = operator.shape();
 
+        assert_eq!(dim[0], eko_info.grid_axes.mur2_grid.len());
         assert_eq!(dim[0], eko_info.grid_axes.muf2_grid.len());
         assert_eq!(dim[1], eko_info.target_pids.len());
         assert_eq!(dim[3], eko_info.grid_axes.pids.len());
@@ -1482,6 +1513,49 @@ impl Grid {
             "[{elapsed_precise}] {bar:50.cyan/blue} {pos:>7}/{len:7} - ETA: {eta_precise} {msg}",
         ));
 
+        // which (tgt_pid, src_pid) tuples are non-zero in general?
+        let non_zero_pid_indices: Vec<_> = (0..operator.dim().0)
+            .cartesian_product(0..operator.dim().1)
+            .filter(|&(tgt_pid_idx, src_pid_idx)| {
+                operator
+                    .slice(s![tgt_pid_idx, src_pid_idx, .., .., ..])
+                    .iter()
+                    .any(|&value| value != 0.0)
+            })
+            .collect();
+
+        // the x1 and x2 grid values are the same across all non-zero subgrids
+        let first_non_zero_grid = self
+            .subgrids
+            .iter()
+            .find(|subgrid| !subgrid.is_empty())
+            .unwrap_or_else(|| unreachable!());
+        // source x1/x2 grid might differ and be differently sorted than the operator
+        let x1_grid: Vec<_> = first_non_zero_grid
+            .x1_grid()
+            .iter()
+            .map(|x| {
+                eko_info
+                    .grid_axes
+                    .x_grid
+                    .iter()
+                    .position(|xi| approx_eq!(f64, *xi, *x, ulps = 64))
+                    .unwrap_or_else(|| unreachable!())
+            })
+            .collect();
+        let x2_grid: Vec<_> = first_non_zero_grid
+            .x2_grid()
+            .iter()
+            .map(|x| {
+                eko_info
+                    .grid_axes
+                    .x_grid
+                    .iter()
+                    .position(|xi| approx_eq!(f64, *xi, *x, ulps = 64))
+                    .unwrap_or_else(|| unreachable!())
+            })
+            .collect();
+
         // iterate over all bins, which are mapped one-to-one from the target to the source grid
         for bin in 0..self.bin_info().bins() {
             // iterate over the source grid luminosities
@@ -1530,32 +1604,6 @@ impl Grid {
 
                     let src_subgrid = &self.subgrids[[order, bin, src_lumi]];
 
-                    // source x1/x2 grid might differ and be differently sorted than the operator
-                    let x1_grid: Vec<_> = src_subgrid
-                        .x1_grid()
-                        .iter()
-                        .map(|x| {
-                            eko_info
-                                .grid_axes
-                                .x_grid
-                                .iter()
-                                .position(|xi| approx_eq!(f64, *xi, *x, ulps = 64))
-                                .unwrap_or_else(|| unreachable!())
-                        })
-                        .collect();
-                    let x2_grid: Vec<_> = src_subgrid
-                        .x2_grid()
-                        .iter()
-                        .map(|x| {
-                            eko_info
-                                .grid_axes
-                                .x_grid
-                                .iter()
-                                .position(|xi| approx_eq!(f64, *xi, *x, ulps = 64))
-                                .unwrap_or_else(|| unreachable!())
-                        })
-                        .collect();
-
                     for ((iq2, ix1, ix2), &value) in src_subgrid.iter() {
                         let scale = src_subgrid.mu2_grid()[iq2].fac;
                         let src_iq2 = src_array_q2_grid
@@ -1564,15 +1612,15 @@ impl Grid {
                             .unwrap();
                         let als_iq2 = eko_info
                             .grid_axes
-                            .muf2_grid
+                            .mur2_grid
                             .iter()
                             .position(|&q2| {
                                 approx_eq!(f64, q2, eko_info.xir * eko_info.xir * scale, ulps = 64)
                             })
                             .unwrap_or_else(|| {
                                 panic!(
-                                    "Couldn't find q2: {:?} with xir: {:?} and muf2_grid: {:?}",
-                                    scale, eko_info.xir, eko_info.grid_axes.muf2_grid
+                                    "Couldn't find mur2: {:?} with xir: {:?} and mur2_grid: {:?}",
+                                    scale, eko_info.xir, eko_info.grid_axes.mur2_grid
                                 )
                             });
 
@@ -1602,8 +1650,15 @@ impl Grid {
                             .grid_axes
                             .muf2_grid
                             .iter()
-                            .position(|&q2| approx_eq!(f64, q2, src_q2, ulps = 64))
-                            .unwrap()
+                            .position(|&q2| {
+                                approx_eq!(f64, q2, eko_info.xif * eko_info.xif * src_q2, ulps = 64)
+                            })
+                            .unwrap_or_else(|| {
+                                panic!(
+                                    "Couldn't find muf2: {:?} with xif: {:?} and muf2_grid: {:?}",
+                                    src_q2, eko_info.xif, eko_info.grid_axes.muf2_grid
+                                )
+                            })
                     })
                     .collect();
                 // Iterate target lumis
@@ -1648,6 +1703,20 @@ impl Grid {
                             0
                         };
 
+                        // if `op1` and `op2` below are zero there's no work to do
+                        // TODO: ideally we change the for loops instead of vetoing here
+                        if (has_pdf1
+                            && !non_zero_pid_indices
+                                .iter()
+                                .any(|&tuple| tuple == (tgt_pid1_idx, src_pid1_idx)))
+                            || (has_pdf2
+                                && !non_zero_pid_indices
+                                    .iter()
+                                    .any(|&tuple| tuple == (tgt_pid2_idx, src_pid2_idx)))
+                        {
+                            continue;
+                        }
+
                         // create target subgrid
                         let mut tgt_array =
                             SparseArray3::new(1, tgt_x1_grid.len(), tgt_x2_grid.len());
@@ -2280,6 +2349,7 @@ mod tests {
     // TODO: properly test axes returned
 
     fn simple_grid() -> (Grid, GridAxes) {
+        let mur2_grid = vec![20.];
         let muf2_grid = vec![20.];
         let x_grid = vec![0.1, 0.5, 1.];
 
@@ -2320,6 +2390,7 @@ mod tests {
             GridAxes {
                 x_grid,
                 pids,
+                mur2_grid,
                 muf2_grid,
             },
         )
@@ -2331,6 +2402,7 @@ mod tests {
 
         let ret_axes = grid.axes().unwrap();
         assert_eq!(ret_axes.x_grid, axes.x_grid);
+        assert_eq!(ret_axes.mur2_grid, axes.mur2_grid);
         assert_eq!(ret_axes.muf2_grid, axes.muf2_grid);
         assert_eq!(ret_axes.pids, vec![]);
     }
@@ -2353,6 +2425,7 @@ mod tests {
             grid_axes: GridAxes {
                 x_grid: axes.x_grid.clone(),
                 pids: axes.pids.clone(),
+                mur2_grid: axes.mur2_grid.clone(),
                 muf2_grid: axes.muf2_grid.clone(),
             },
             lumi_id_types,
